﻿package stego.core
{
    import flash.display.Bitmap;
    import flash.display.DisplayObject;
    import flash.display.DisplayObjectContainer;
    import flash.display.Shape;
    import flash.display.Sprite;
    import flash.display.Stage;
    import flash.display.StageScaleMode;
    import flash.events.Event;
    import flash.events.KeyboardEvent;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.core.GameInterface;
    import stego.core.GameObject;
    import stego.debug.DebugWidget;
    import stego.InputUtil;
    import stego.Settings;
    import stego.util.MathUtil;
    import stego.util.PhysicsUtil;
    import stego.screen.Pause;
    import stego.core.GameState;
    import stego.Util;
    import stego.util.CheatUtil;
    import stego.util.music.MusicUtil;
    import stego.util.SceneLayers;
    import stego.util.TimeUtil;
    import stego.widget.MusicCredits;
    import stego.widget.TextField;

    /**
     * Base class that you will be extending to create your own games.
     * @author Mike Meyer
     */
    public class BaseGame extends Sprite implements GameInterface
    {
        protected var mGameState:GameState;
        private var mNextGameState:GameState;
        private var mIsPaused:Boolean;
        private static const CAMERA_TIGHTNESS:Number = 0.125;

        /// widget for music credits
        protected var mMusicCredits:MusicCredits;

        private var _cameraPosition:Point = new Point(0,0);
        public function get CameraPosition():Point { return _cameraPosition; }

        public function get CameraOffset():Point
        {
            return new Point(
                CameraPosition.x - 0.5 * StageWidth,
                CameraPosition.y - 0.5 * StageHeight );
        }

        /// see how much the game is zoomed in
        public function get Zoom():Number { return mGameHolder.scaleX; }

        /// set how much the game is zoomed in
        public function set Zoom(zoom:Number):void { mGameHolder.scaleX = zoom; mGameHolder.scaleY = zoom; }

        private var mPauseScreen:Sprite;
        private var mGameHolder:DisplayObjectContainer;

        private var _cheatUtil:CheatUtil;
        public function get Cheats():CheatUtil { return _cheatUtil; }

        private var _musicUtil:MusicUtil;
        public function get Music():MusicUtil { return _musicUtil; }

        private var _layers:SceneLayers;
        public function get Layers():SceneLayers { return _layers; }

        private var _time:TimeUtil;
        public function get Time():TimeUtil { return _time; }

        private var _physics:PhysicsUtil;
        public function get Physics():PhysicsUtil { return _physics; }

        private var _debugger:DebugWidget;
        public function get Debug():DebugWidget { return _debugger; }

        private var _cameraShake:Number = 0;
        public function get CameraShake():Number { return _cameraShake; }
        public function set CameraShake(value:Number):void { _cameraShake = value;  }

        private var _cameraFollowTarget:GameObject = null;
        public function get CameraFollowTarget():GameObject { return _cameraFollowTarget; }
        public function set CameraFollowTarget(value:GameObject):void { _cameraFollowTarget = value;  }

        private var _cameraSpeed:Number = 1000;
        public function get CameraSpeed():Number { return _cameraSpeed; }
        public function set CameraSpeed(value:Number):void { _cameraSpeed = value; }

        private var _inputUtil:InputUtil;
        public function get inputUtil():InputUtil { return _inputUtil; }
        public function set inputUtil(value:InputUtil):void { _inputUtil = value; }

        public function get StageRect():Rectangle { return new Rectangle(0, 0, StageWidth, StageHeight); }

        // shake looks much better if the shaking tends to
        // go back and forth instead of actually random
        private var _cameraShakeDir:Number = 0;
        private static const MIN_SHAKE_ROT:Number = 0.67 * Math.PI;
        private static const MAX_SHAKE_ROT:Number = 1.33 * Math.PI;

        private var _stageWidth:Number;
        public function get StageWidth():Number { return _stageWidth; }
        private var _stageHeight:Number;
        public function get StageHeight():Number { return _stageHeight; }

        /**
         * Automagically called by the constructor for the initial init,
         * but if you shut it down and want to restart, you'll need to call it yourself.
         *
         * override (but don't forget to call super.Init() to customize your game
         */
        public function Init(stageWidth:Number, stageHeight:Number):void
        {
            if (null == stage)
            {
                throw new Error("Called init before adding game to the stage!  Please add it first!");
            }

            _stageWidth = stageWidth;
            _stageHeight = stageHeight;

            _cameraPosition = new Point(
                0.5 * _stageWidth,
                0.5 * _stageHeight );

            _cheatUtil = new CheatUtil();
            Cheats.Init(stage);

            _musicUtil = new MusicUtil();
            Music.Init();

            _time = new TimeUtil();
            Time.Init();

            _physics = new PhysicsUtil();

            CameraFollowTarget = null;

            mIsPaused = false;

            _inputUtil = new InputUtil(stage);

            addEventListener(Event.ENTER_FRAME, HandleEnterFrame);

            SetupScene();

            mNextGameState = null;

            stage.addEventListener(Event.DEACTIVATE, _Pause);
            stage.addEventListener(Event.ACTIVATE, _Unpause);
        }

        /**
         * add the persistent display objects that stick around even between levels
         */
        private function SetupScene():void
        {
            // holder for the game itself (gets moved to scroll the camera)
            mGameHolder = new Sprite();
            this.addChild(mGameHolder);

            _layers = new SceneLayers();
            Layers.Init(mGameHolder);

            _debugger = new DebugWidget();
            this.addChild(_debugger);

            // display for the music credits
            mMusicCredits = new MusicCredits(Music, StageWidth);
            mMusicCredits.Init();
            this.addChild(mMusicCredits);

            CreatePauseIndicator();
        }

        /**
         * create the pause screen
         */
        protected function CreatePauseIndicator():void
        {
            mPauseScreen = new stego.screen.Pause(StageWidth, StageHeight);
            this.addChild(mPauseScreen);
        }

        protected function HandleEnterFrame(e:Event = null):void
        {
            Time.TimeDelta = (1.0 / stage.frameRate) * Time.TimeScale;

            ScriptChange();

            if (!mIsPaused)
            {
                Update();
                Integrate();
                PreDraw();
            }
            inputUtil.Update();
        }

        protected function Update():void
        {
            if (null != mGameState)
            {
                mGameState.Update();
            }
            Layers.Update(Time.TimeDelta);
            Physics.Update(Layers, Time.TimeDelta);
            mMusicCredits.Update(Time.TimeDelta);

            // camera shake
            _cameraShakeDir += MathUtil.Lerp( MIN_SHAKE_ROT, MAX_SHAKE_ROT, Math.random() );
            _cameraShakeDir = MathUtil.Wrap( _cameraShakeDir, 0, 2.0 * Math.PI);
            CameraUpdate();
        }

        /**
         * update the movieclips to reflect the new camera position
         */
        protected function CameraUpdate():void
        {
            mGameHolder.x = (0.5 * StageWidth) - CameraPosition.x + (CameraShake * Math.cos(_cameraShakeDir));
            mGameHolder.y = (0.5 * StageHeight) - CameraPosition.y + (CameraShake * Math.sin(_cameraShakeDir));
        }

        protected function Integrate():void
        {
            Layers.Integrate(Time.TimeDelta);
        }

        protected function PreDraw():void
        {
            if (null != mGameState)
            {
                mGameState.PreDraw();
            }
            Layers.PreDraw();
        }

        /// Called on each Beat of the game.
        protected function Beat(e:Event = null):void
        {
            var beat:Number = mGameState.GetCurBeat();

            Layers.Beat(beat);
        }

        protected function OffBeat(e:Event = null):void
        {
            var beat:Number = mGameState.GetCurBeat();

            Layers.OffBeat(beat);
        }
        protected function QuarterBeat(e:Event = null):void
        {
            var beat:Number = mGameState.GetCurQuarterBeat();

            Layers.QuarterBeat(beat);
        }

        /**
         * freeze time and show the pause screen
         */
        public function Pause():void
        {
            mIsPaused = true;
            mPauseScreen.visible = true;
            Music.Pause();
        }

        /**
         * light wrapper for using Pause as an event callback
         */
        private function _Pause(e:Event):void
        {
            Pause();
        }

        /**
         * resume after pausing
         */
        public function Unpause():void
        {
            mIsPaused = false;
            mPauseScreen.visible = false;
            Music.Resume();
        }

        /**
         * light wrapper for using Pause as an event callback
         */
        private function _Unpause(e:Event):void
        {
            Unpause();
        }

        /**
         * if a script change was queued up, perform the actual switch now
         */
        private function ScriptChange():void
        {
            if (mNextGameState != null)
            {
                try
                {
                    EndScript(mGameState);
                    var temp:GameState = mNextGameState; // careful! something triggered by Begin might change mNextGameState!
                    mNextGameState = null;

                    BeginScript(temp);
                    mGameState = temp;
                }
                catch (e:Error)
                {
                    // don't try to do that again
                    mNextGameState = null;
                    // rethrow
                    throw e;
                }
            }
        }

        /**
         * shut it down
         */
        private function EndScript(script:GameState):void
        {
            if (script != null)
            {
                script.removeEventListener(GameState.EVENT_BEAT, Beat);
                script.removeEventListener(GameState.EVENT_OFFBEAT, OffBeat);
                script.removeEventListener(GameState.EVENT_QUARTERBEAT, QuarterBeat);
                script.End();
            }
        }

        /**
         *
         */
        private function BeginScript(script:GameState):void
        {
            if (script != null)
            {
                script.addEventListener(GameState.EVENT_BEAT, Beat);
                script.addEventListener(GameState.EVENT_OFFBEAT, OffBeat);
                script.addEventListener(GameState.EVENT_QUARTERBEAT, QuarterBeat);
                script.Begin(this);
            }
        }

        /**
         * set a script to switch to once the current beat/update is finished
         */
        public function ChangeGameState(nextScript:GameState):void
        {
            mNextGameState = nextScript;
        }

        /**
         * go back to the starting position
         */
        public function ResetScrolling():void
        {
            mGameHolder.x = 0;
            mGameHolder.y = 0;
        }

        /**
         * Clean up
         */
        public function Shutdown():void
        {
            CameraFollowTarget = null;
            mGameState = null;
            mNextGameState = null;

            removeEventListener(Event.ENTER_FRAME, HandleEnterFrame);
            stage.removeEventListener(Event.DEACTIVATE, _Pause);
            stage.removeEventListener(Event.ACTIVATE, _Unpause);

            if (null != mGameHolder)
            {
                this.removeChild(mGameHolder);
                mGameHolder = null;
            }

            if (null != mPauseScreen)
            {
                this.removeChild(mPauseScreen);
                mPauseScreen = null;
            }

            if (null != this.Cheats)
            {
                this.Cheats.Shutdown();
                _cheatUtil = null;
            }

            if (null != this.Music)
            {
                this.Music.Shutdown();
                _musicUtil = null;
            }

            if (null != this.Layers)
            {
                this.Layers.Shutdown();
                _layers = null;
            }

            if (null != Time)
            {
                Time.Shutdown();
                _time = null;
            }
        }


        /**
         * do the scrolling to center the camera on the given location
         */
        public function SetCameraPosition(x:Number, y:Number):void
        {
            CameraPosition.x = (0.5 * StageWidth) + x;
            CameraPosition.y = (0.5 * StageHeight) + y;
            CameraUpdate();
        }
    }
}
